declare function Brand<B extends string, A extends RuntypeBase>(brand: B, entity: A): any;

declare interface Brand<B extends string, A extends RuntypeBase> extends Runtype<Static<A> & RuntypeBrand<B>> {
    tag: 'brand';
    brand: B;
    entity: A;
}

declare type Case<T extends RuntypeBase, Result> = (v: Static<T>) => Result;

/**
 * @module utils
 */
/**
 * The class containing 8-bit value and indices information
 */
declare abstract class Codeword {
    /** The index through all blocks before interleaving */
    protected pre_interleave_index: number;
    /** The index of a block in an array of blocks */
    protected block_index: number;
    /** The index in a block */
    protected index: number;
    /** The index through all blocks after interleaving */
    protected post_interleave_index: number;
    /** The 8-bit value */
    protected readonly value: number;
    /**
     * Create a new Codeword instance
     *
     * @param value - The 8-bit value represented by an integer
     * @protected
     */
    protected constructor(value: number);
    /**
     * Set the index of a block
     *
     * @param value - The index of a block
     */
    setBlockIndex(value: number): void;
    /**
     * Set the index of itself in a block
     *
     * @param value - The index of itself
     */
    setIndex(value: number): void;
    /**
     * Set the index of itself through all blocks before interleaving
     *
     * @param value - The index through all blocks before interleaving
     */
    setPreInterleaveIndex(value: number): void;
    /**
     * Set the index of itself through all blocks after interleaving
     *
     * @param value - The index through all blocks after interleaving
     */
    setPostInterleaveIndex(value: number): void;
    /** Return the 8-bit value */
    getValue(): number;
}

declare function Constraint<A extends RuntypeBase, T extends Static<A> = Static<A>, K = unknown>(underlying: A, constraint: ConstraintCheck<A>, options?: {
    name?: string;
    args?: K;
}): Constraint<A, T, K>;

declare interface Constraint<A extends RuntypeBase, T extends Static<A> = Static<A>, K = unknown> extends Runtype<T> {
    tag: 'constraint';
    underlying: A;
    constraint(x: Static<A>): boolean | string;
    name?: string;
    args?: K;
}

declare type ConstraintCheck<A extends RuntypeBase> = (x: Static<A>) => boolean | string;

declare interface Constructor<V> {
    new (...args: any[]): V;
}

/**
 * A detailed object enumerating where the validation failed exactly.
 */
declare type Details = (string | Details)[] | {
    [key in string | number | symbol]: string | Details;
};

/**
 * A Union of error correction level in string literal
 * for type aliasing.
 * External library {@link https://github.com/pelotom/runtypes | runtypes} used here.
 */
declare const Ecl: Union<[Literal<"LOW">, Literal<"MEDIUM">, Literal<"QUARTILE">, Literal<"HIGH">]>;

declare type Ecl = Static<typeof Ecl>;

/**
 * A predefined error code indicating what type of failure has occured.
 */
declare type Failcode = typeof Failcode[keyof typeof Failcode];

declare const Failcode: {
    /** The type of the received primitive value is incompatible with expected one. */
    readonly TYPE_INCORRECT: "TYPE_INCORRECT";
    /** The received primitive value is incorrect. */
    readonly VALUE_INCORRECT: "VALUE_INCORRECT";
    /** The key of the property is incorrect. */
    readonly KEY_INCORRECT: "KEY_INCORRECT";
    /** One or more elements or properties of the received object are incorrect. */
    readonly CONTENT_INCORRECT: "CONTENT_INCORRECT";
    /** One or more arguments passed to the function is incorrect. */
    readonly ARGUMENT_INCORRECT: "ARGUMENT_INCORRECT";
    /** The value returned by the function is incorrect. */
    readonly RETURN_INCORRECT: "RETURN_INCORRECT";
    /** The received value does not fulfill the constraint. */
    readonly CONSTRAINT_FAILED: "CONSTRAINT_FAILED";
    /** The property must be present but missing. */
    readonly PROPERTY_MISSING: "PROPERTY_MISSING";
    /** The property must not be present but present. */
    readonly PROPERTY_PRESENT: "PROPERTY_PRESENT";
    /** The value must not be present but present. */
    readonly NOTHING_EXPECTED: "NOTHING_EXPECTED";
};

/**
 * A failed validation result.
 */
declare type Failure = {
    /**
     * A tag indicating failure.
     */
    success: false;
    /**
     * An error code assigned to this type of error.
     */
    code: Failcode;
    /**
     * A message indicating the reason why the validation failed.
     */
    message: string;
    /**
     * A detailed object enumerating where the validation failed exactly.
     */
    details?: Details;
};

/**
 * Wrap up all steps for generating the QR code
 *
 * @param text - The string that should be encoded to QR code
 * @param options - Options for how to generate the QR code
 */
export declare function generate(text: string, options?: GeneratorOptions): QR;

declare interface GeneratorOptions {
    /** The **minimal** version of QR code, default is 1 */
    version?: Version;
    /** The error correction level of QR code, default is "LOW" */
    ecl?: Ecl;
    /** The choice of forcing to use the minimal version, default is false */
    forced?: boolean;
    /** The choice of mask patterns, default is -1 */
    mask?: Mask;
}

/**
 * Construct an intersection runtype from runtypes for its alternatives.
 */
declare function Intersect<A extends readonly [RuntypeBase, ...RuntypeBase[]]>(...intersectees: A): Intersect<A>;

declare interface Intersect<A extends readonly [RuntypeBase, ...RuntypeBase[]]> extends Runtype<{
    [K in keyof A]: A[K] extends RuntypeBase ? (parameter: Static<A[K]>) => any : unknown;
}[number] extends (k: infer I) => void ? I : never> {
    tag: 'intersect';
    intersectees: A;
}

/**
 * Construct a runtype for a type literal.
 */
declare function Literal<A extends LiteralBase>(valueBase: A): Literal<A>;

declare interface Literal<A extends LiteralBase> extends Runtype<A> {
    tag: 'literal';
    value: A;
}

/**
 * The super type of all literal types.
 */
declare type LiteralBase = undefined | null | boolean | number | bigint | string;

/**
 * A number between -1 and 7 (including)
 * for type aliasing.
 * External library {@link https://github.com/pelotom/runtypes | runtypes} used here.
 */
declare const Mask: Constraint<Number_2, number, unknown>;

declare type Mask = Static<typeof Mask>;

declare interface Match<A extends readonly [RuntypeBase, ...RuntypeBase[]]> {
    <Z>(...a: {
        [K in keyof A]: A[K] extends RuntypeBase ? Case<A[K], Z> : never;
    }): Matcher<A, Z>;
}

declare type Matcher<A extends readonly [RuntypeBase, ...RuntypeBase[]], Z> = (x: {
    [K in keyof A]: A[K] extends RuntypeBase<infer Type> ? Type : unknown;
}[number]) => Z;

/**
 * The Class representing a square in QR code
 */
declare class Module {
    /** TRUE for dark, FALSE for light */
    protected color: boolean;
    /**
     * Create a new Module instance
     *
     * @param color - The binary data represented by boolean
     */
    constructor(color?: boolean);
    /**
     * Set the color to a new value
     *
     * @param value - The new color
     */
    setColor(value: boolean): void;
    /** Return the color */
    getColor(): boolean;
    /** Deep copy itself */
    copy(): Module;
}

/**
 * An alias for Literal(null).
 */
declare const Null: Literal<null>;

declare interface Number_2 extends Runtype<number> {
    tag: 'number';
}

/**
 * Validates that a value is a number.
 */
declare const Number_2: Number_2;

/**
 * Validates optional value.
 */
declare function Optional<R extends RuntypeBase>(runtype: R): Optional<R>;

declare interface Optional<R extends RuntypeBase> extends Runtype<Static<R> | undefined> {
    tag: 'optional';
    underlying: R;
}

/**
 * A 2D position represented by a tuple of integer
 * for type aliasing.
 * External library {@link https://github.com/pelotom/runtypes | runtypes} used here.
 */
declare const Position: Tuple<[Number_2, Number_2]>;

declare type Position = Static<typeof Position>;

/**
 * The class for containing all information to generate a QR code
 */
declare class QR {
    /** The version of QR code*/
    private readonly version;
    /** The error correction level */
    private readonly ecl;
    /** The size of QR code according to the version */
    private readonly size;
    /** All data represented by a 2D array of module objects */
    private modules;
    /**
     * Create a new QR instance
     *
     * @param version - The version of QR code
     * @param ecl - The error correction level
     * @param modules - The initial data for copying
     */
    constructor(version: Version, ecl: Ecl, modules?: Array<Array<Module>>);
    /**
     * Get the specific bit of the given *\<bits\>* at *\<index\>*
     * from right to left
     *
     * @param bits
     * @param index
     * @private
     */
    private static getBitAt;
    /** Return the size */
    getSize(): number;
    /** Return the data */
    getModules(): Array<Array<Module>>;
    /** Deep copy itself */
    copy(): QR;
    /** Stringify the data */
    toString(): string;
    /** Initializing by drawing fixed patterns */
    initialize(): void;
    /** Draw the timing patterns */
    private drawTimingPatterns;
    /** Draw the finder patterns */
    private drawFinderPatterns;
    /** Draw the separator patterns */
    private drawSeparatorPatterns;
    /** Draw the alignment patterns */
    private drawAlignmentPatterns;
    /** Draw the dark pattern */
    private drawDarkPattern;
    /** Draw the format patterns */
    drawFormatPatterns(mask?: Mask): void;
    /** Draw the version information patterns */
    private drawVersionInfoPatterns;
    /** Generate an array of positions for placing data */
    generateDataPath(): Array<Position>;
    /**
     * Place the given *\<data\>* into the class through *\<path\>*
     *
     * @param data  The array of codeword
     * @param path  The array of position
     */
    placeCodeword(data: Array<Codeword>, path: Array<Position>): void;
    /**
     * Generate a mask specified by the given *\<which\>*
     *
     * @param which
     */
    generateMask(which: Mask): QR;
    /** Generate all kinds of masks */
    generateAllMasks(): Array<QR>;
    /**
     * Apply the given *\<mask\>* to the data in the class
     *
     * @param mask  The QR object
     */
    applyMask(mask: QR): void;
    /** Compute the penalty for the class */
    computePenalty(): number;
    /** Add 4 rows of light modules */
    private generateQuietRows;
    /** Add a quiet zone - a 4-module-wide area of light modules around the QR code */
    extend(): void;
}

declare type Reflect_2 = ({
    tag: 'unknown';
} & Runtype) | ({
    tag: 'never';
} & Runtype<never>) | ({
    tag: 'void';
} & Runtype<void>) | ({
    tag: 'boolean';
} & Runtype<boolean>) | ({
    tag: 'number';
} & Runtype<number>) | ({
    tag: 'bigint';
} & Runtype<bigint>) | ({
    tag: 'string';
} & Runtype<string>) | ({
    tag: 'symbol';
    key: string | undefined;
} & Runtype<symbol>) | ({
    tag: 'symbol';
    (key: string | undefined): Runtype<symbol>;
} & Runtype<symbol>) | ({
    tag: 'literal';
    value: LiteralBase;
} & Runtype<LiteralBase>) | ({
    tag: 'template';
    strings: string[];
    runtypes: Runtype<LiteralBase>[];
} & Runtype<string>) | ({
    tag: 'array';
    element: Reflect_2;
    isReadonly: boolean;
} & Runtype<ReadonlyArray<unknown>>) | ({
    tag: 'record';
    fields: {
        [_: string]: Reflect_2;
    };
    isPartial: boolean;
    isReadonly: boolean;
} & Runtype<{
    readonly [_ in string]: unknown;
}>) | ({
    tag: 'dictionary';
    key: 'string' | 'number' | 'symbol';
    value: Reflect_2;
} & Runtype<{
    [_: string]: unknown;
}>) | ({
    tag: 'tuple';
    components: Reflect_2[];
} & Runtype<unknown[]>) | ({
    tag: 'union';
    alternatives: Reflect_2[];
} & Runtype) | ({
    tag: 'intersect';
    intersectees: Reflect_2[];
} & Runtype) | ({
    tag: 'optional';
    underlying: Reflect_2;
} & Runtype) | ({
    tag: 'function';
} & Runtype<(...args: any[]) => any>) | ({
    tag: 'constraint';
    underlying: Reflect_2;
    constraint: ConstraintCheck<Runtype<never>>;
    args?: any;
    name?: string;
} & Runtype) | ({
    tag: 'instanceof';
    ctor: Constructor<unknown>;
} & Runtype) | ({
    tag: 'brand';
    brand: string;
    entity: Reflect_2;
} & Runtype);

/**
 * Render QR code on HTML canvas
 * @param code - the QR code
 * @param id - the id of the canvas element
 * @param options - options for how to render the QR code
 */
export declare function renderOnCanvas(code: QR, id: string, options?: RenderOptions): HTMLCanvasElement;

declare interface RenderOptions {
    /**
     * The color choice of dark module/pixels,
     * suppose to be a valid hex color number string,
     * default is #000000 - stands for black
     */
    c1?: string;
    /**
     * The color choice of light module/pixels,
     * suppose to be a valid hex color number string,
     * default is #FFFFFF - stands for white
     */
    c2?: string;
    /** The size/width of a module/square, default is 10 */
    size?: number;
}

/**
 * The result of a type validation.
 */
declare type Result<T> = Success<T> | Failure;

/**
 * A runtype determines at runtime whether a value conforms to a type specification.
 */
declare interface Runtype<A = unknown> extends RuntypeBase<A> {
    /**
     * Union this Runtype with another.
     */
    Or<B extends RuntypeBase>(B: B): Union<[this, B]>;
    /**
     * Intersect this Runtype with another.
     */
    And<B extends RuntypeBase>(B: B): Intersect<[this, B]>;
    /**
     * Optionalize this Runtype.
     */
    optional(): Optional<this>;
    /**
     * Union this Runtype with `Null`.
     */
    nullable(): Union<[this, typeof Null]>;
    /**
     * Use an arbitrary constraint function to validate a runtype, and optionally
     * to change its name and/or its static type.
     *
     * @template T - Optionally override the static type of the resulting runtype
     * @param {(x: Static<this>) => boolean | string} constraint - Custom function
     * that returns `true` if the constraint is satisfied, `false` or a custom
     * error message if not.
     * @param [options]
     * @param {string} [options.name] - allows setting the name of this
     * constrained runtype, which is helpful in reflection or diagnostic
     * use-cases.
     */
    withConstraint<T extends Static<this>, K = unknown>(constraint: ConstraintCheck<this>, options?: {
        name?: string;
        args?: K;
    }): Constraint<this, T, K>;
    /**
     * Helper function to convert an underlying Runtype into another static type
     * via a type guard function.  The static type of the runtype is inferred from
     * the type of the guard function.
     *
     * @template T - Typically inferred from the return type of the type guard
     * function, so usually not needed to specify manually.
     * @param {(x: Static<this>) => x is T} guard - Type guard function (see
     * https://www.typescriptlang.org/docs/handbook/advanced-types.html#user-defined-type-guards)
     *
     * @param [options]
     * @param {string} [options.name] - allows setting the name of this
     * constrained runtype, which is helpful in reflection or diagnostic
     * use-cases.
     */
    withGuard<T extends Static<this>, K = unknown>(guard: (x: Static<this>) => x is T, options?: {
        name?: string;
        args?: K;
    }): Constraint<this, T, K>;
    /**
     * Adds a brand to the type.
     */
    withBrand<B extends string>(brand: B): Brand<B, this>;
}

/**
 * A runtype determines at runtime whether a value conforms to a type specification.
 */
declare interface RuntypeBase<A = unknown> {
    /**
     * Verifies that a value conforms to this runtype. When given a value that does
     * not conform to the runtype, throws an exception.
     */
    assert(x: any): asserts x is A;
    /**
     * Verifies that a value conforms to this runtype. If so, returns the same value,
     * statically typed. Otherwise throws an exception.
     */
    check(x: any): A;
    /**
     * Validates that a value conforms to this type, and returns a result indicating
     * success or failure (does not throw).
     */
    validate(x: any): Result<A>;
    /**
     * A type guard for this runtype.
     */
    guard(x: any): x is A;
    /**
     * Convert this to a Reflect, capable of introspecting the structure of the type.
     */
    readonly reflect: Reflect_2;
    readonly _falseWitness: A;
    readonly [RuntypeSymbol]: true;
}

declare interface RuntypeBrand<B extends string> {
    [RuntypeName]: B;
}

declare const RuntypeName: unique symbol;

declare const RuntypeSymbol: unique symbol;

/**
 * Obtains the static type associated with a Runtype.
 */
declare type Static<A extends RuntypeBase> = A['_falseWitness'];

/**
 * A successful validation result.
 */
declare type Success<T> = {
    /**
     * A tag indicating success.
     */
    success: true;
    /**
     * The original value, cast to its validated type.
     */
    value: T;
};

/**
 * Construct a tuple runtype from runtypes for each of its elements.
 */
declare function Tuple<T extends readonly RuntypeBase[]>(...components: T): Tuple<T>;

declare interface Tuple<A extends readonly RuntypeBase[]> extends Runtype<{
    [K in keyof A]: A[K] extends RuntypeBase ? Static<A[K]> : unknown;
}> {
    tag: 'tuple';
    components: A;
}

/**
 * Construct a union runtype from runtypes for its alternatives.
 */
declare function Union<T extends readonly [RuntypeBase, ...RuntypeBase[]]>(...alternatives: T): Union<T>;

declare interface Union<A extends readonly [RuntypeBase, ...RuntypeBase[]]> extends Runtype<{
    [K in keyof A]: A[K] extends RuntypeBase ? Static<A[K]> : unknown;
}[number]> {
    tag: 'union';
    alternatives: A;
    match: Match<A>;
}

/**
 * A number between 1 and 40 (including)
 * for type aliasing.
 * External library {@link https://github.com/pelotom/runtypes | runtypes} used here.
 */
declare const Version: Constraint<Number_2, number, unknown>;

declare type Version = Static<typeof Version>;

export { }
